package com.agilecontrol.nea.core.web.shell;

import argparser.ArgParser;
import argparser.IntHolder;
import argparser.StringHolder;
import com.agilecontrol.csvreader.CsvWriter;
import com.agilecontrol.nea.core.db.oracle.QDate;
import com.agilecontrol.nea.core.io.PluginLifecycleListener;
import com.agilecontrol.nea.core.query.QueryEngine;
import com.agilecontrol.nea.core.query.QueryRequestImpl;
import com.agilecontrol.nea.core.query.QuerySession;
import com.agilecontrol.nea.core.query.QueryUtils;
import com.agilecontrol.nea.core.report.ReportUtils;
import com.agilecontrol.nea.core.schema.Column;
import com.agilecontrol.nea.core.schema.Table;
import com.agilecontrol.nea.core.schema.TableManager;
import com.agilecontrol.nea.core.util.ConfigValues;
import com.agilecontrol.nea.core.web.shell.AbstractShellCmd;
import com.agilecontrol.nea.core.web.shell.Shell;
import com.agilecontrol.nea.util.NDSRuntimeException;
import com.agilecontrol.nea.util.Tools;
import com.agilecontrol.nea.util.Validator;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import javax.servlet.http.HttpServletRequest;
import org.json.JSONArray;
import org.json.JSONObject;

@Shell(
   alias = "tc",
   admin = true,
   help = "tc -c start|stop|status|done -t [间隔分钟数] 时间胶囊TimeCapsule, 订单明细定时备份"
)
public class TimeCapsule extends AbstractShellCmd implements PluginLifecycleListener {

   private static TimeCapsule.Backup backupInstance = null;
   private static final String RUNNING_MARK_FILENAME = "run.tc";
   private static final String FOITEM_SQL = "select to_char(sysdate,\'YYYY-MM-DD HH24:MI:SS\')||\'.000\' ts, i.m_product_id pdtid, fu.user_id userid,sum(i.qty) qty, sum(i.amt_list) amtlist from b_foitem i, b_fo fo, b_funit fu where i.b_fo_id=fo.id and fu.id=fo.b_funit_id and fu.fotype=\'STANDARD\' group by i.m_product_id, fu.user_id";


   private void cleanFolder() {
      try {
         String tx = QueryEngine.getInstance().doQueryString("select to_char(min(creationdate),\'YYYYMMDDHH24MISS\') from b_foitem", new Object[0]);
         long minCreationDate = System.currentTimeMillis();
         if(Validator.isNotNull(tx)) {
            long folder = Tools.getLong(tx, -1L);
            if(folder > 0L) {
               SimpleDateFormat fmt = new SimpleDateFormat("yyyyMMddHHmmss");
               minCreationDate = fmt.parse(tx).getTime();
            }
         }

         File var10 = new File(ConfigValues.DIR_NEA_ROOT + "/tc");
         if(var10.exists()) {
            File[] var5 = var10.listFiles();
            int var11 = var5.length;

            for(int var7 = 0; var7 < var11; ++var7) {
               File file = var5[var7];
               if(file.isFile() && file.lastModified() < minCreationDate && !file.delete()) {
                  this.logger.warn("Fail to delete " + file.getAbsolutePath());
               }
            }
         }
      } catch (Throwable var9) {
         this.logger.error("Fail to clean folder", var9);
      }

   }

   private String startService(int intevalInMinutes) throws Exception {
      if(backupInstance != null && backupInstance.isRunning()) {
         return "TimeCapsule is running, please stop first";
      } else {
         boolean backupEveryTime = false;
         Connection conn = QueryEngine.getInstance().getConnection();

         try {
            backupEveryTime = Validator.isNotNull(this.getAdSQL("tc_foitem", conn));
         } finally {
            try {
               conn.close();
            } catch (Throwable var10) {
               ;
            }

         }

         this.cleanFolder();
         backupInstance = new TimeCapsule.Backup(intevalInMinutes, backupEveryTime);
         Thread thread = new Thread(backupInstance, "TimeCapsule");
         thread.start();
         this.setRunningMark(true);
         return "TimeCapsule started";
      }
   }

   private int getRunningMark() {
      File folder = new File(ConfigValues.DIR_NEA_ROOT + "/tc");
      File file = new File(folder, "run.tc");
      if(file.exists()) {
         int interVal = 5;

         try {
            interVal = Tools.getInt(Tools.readFile("run.tc", "UTF-8"), 5);
         } catch (IOException var5) {
            this.logger.error("Fail to read content of " + file.getAbsolutePath(), var5);
         }

         return interVal;
      } else {
         return 0;
      }
   }

   private void setRunningMark(boolean running) {
      File folder = new File(ConfigValues.DIR_NEA_ROOT + "/tc");
      if(!folder.exists()) {
         folder.mkdirs();
      }

      if(running) {
         int file = backupInstance == null?5:backupInstance.interval;

         try {
            Tools.writeFile(folder + "/" + "run.tc", String.valueOf(file), "UTF-8");
         } catch (IOException var5) {
            this.logger.error("Fail to create state file run.tc", var5);
         }
      } else {
         File file1 = new File(folder, "run.tc");
         if(file1.exists() && !file1.delete()) {
            this.logger.error("Fail to delete state file run.tc");
         }
      }

   }

   private String stopService(boolean isManual) throws Exception {
      if(isManual) {
         this.setRunningMark(false);
      }

      if(backupInstance == null) {
         return "TimeCapsule is not running";
      } else {
         backupInstance.stop();
         int cnt = 0;
         short maxRetryTimes = 200;

         while(backupInstance.isRunning()) {
            backupInstance.wakeUp();
            Thread.sleep(100L);
            ++cnt;
            if(cnt > maxRetryTimes) {
               break;
            }
         }

         return backupInstance.isRunning()?"Could not stop TimeCapsule, please retry":"TimeCapsule stopped";
      }
   }

   private JSONObject done(HttpServletRequest request) throws Exception {
      String message = null;
      if(backupInstance != null && backupInstance.isRunning()) {
         message = this.stopService(true);
         if(backupInstance.isRunning()) {
            return this.errorObj(message);
         }
      }

      String path = ConfigValues.DIR_NEA_ROOT + "/tc";
      File folder = new File(path);
      if(!folder.exists()) {
         return this.errorObj("Please start at least once");
      } else {
         this.backupProduct();
         this.backupUser();
         this.backupFair();
         ReportUtils ru = new ReportUtils(request);
         String name = ru.getUserName();
         String svrPath = ru.getExportRootPath() + File.separator + ru.getUser().getClientDomain() + File.separator + ru.getUser().getUserId();
         File fp = new File(svrPath);
         if(!fp.exists()) {
            fp.mkdirs();
         }

         String zipFile = svrPath + File.separator + "tc" + ((SimpleDateFormat)QueryUtils.dateNumberFormatter.get()).format(new Date()) + ".zip";
         this.zipFolder((String)path, (String)zipFile);
         File f = new File(zipFile);
         if(f.exists()) {
            JSONObject jo = new JSONObject();
            jo.put("code", 0);
            jo.put("message", "File is created: " + zipFile);
            jo.put("script", "window.location=\'/servlets/binserv/Download?filename=" + f.getName() + "\'");
            return jo;
         } else {
            return this.errorObj("Fail to create " + zipFile);
         }
      }
   }

   private void backupProduct() throws Exception {
      String colFilter = "(m_dim.*|colors)";
      ArrayList colNames = this.getAvailableColumns("i_product", colFilter);
      StringBuilder sql = new StringBuilder("select p.id pdtid,i.name,i.pricelist");
      Iterator var4 = colNames.iterator();

      while(var4.hasNext()) {
         String col = (String)var4.next();
         sql.append(",i.").append(col);
      }

      sql.append(" from i_product i left join m_product p on (p.name=i.name) where i.isactive=\'Y\'");
      this.backupSQL("tc_pdt", sql.toString(), "pdt.csv");
   }

   private void backupUser() throws Exception {
      String colFilter = "(truename|utype|fgroup|by.*|mg.*|area)";
      ArrayList colNames = this.getAvailableColumns("i_user", colFilter);
      Column nameCol = TableManager.getInstance().getColumn("i_user.name");
      String iUserNameCol;
      if(nameCol != null) {
         iUserNameCol = "name";
      } else {
         iUserNameCol = "username";
      }

      StringBuilder sql = new StringBuilder("select usr.id userid, usr.name");
      Iterator var6 = colNames.iterator();

      while(var6.hasNext()) {
         String col = (String)var6.next();
         sql.append(",u.").append(col);
      }

      sql.append(" from i_user u left join users usr on (usr.name=u." + iUserNameCol + ") where u.isactive=\'Y\'");
      this.backupSQL("tc_usr", sql.toString(), "usr.csv");
   }

   private String getAdSQL(String name, Connection conn) throws Exception {
      String sql = "select value from ad_sql where name=?";
      String value = QueryUtils.parseClobOrString(QueryEngine.getInstance().doQueryOne(sql, new Object[]{name}, conn));
      return value;
   }

   private ArrayList getAvailableColumns(String tableName, String colFilter) throws Exception {
      Table table = TableManager.getInstance().getTable(tableName);
      if(table == null) {
         throw new NDSRuntimeException("Not found table " + tableName);
      } else {
         ArrayList selections = new ArrayList();
         QueryRequestImpl query = QueryEngine.getInstance().createRequest((QuerySession)null);
         query.setMainTable(table.getId());
         Iterator rows = table.getAllColumns().iterator();

         while(rows.hasNext()) {
            Column count = (Column)rows.next();
            if(count.getName().toLowerCase().matches(colFilter)) {
               selections.add(count);
               query.addSelection(count.getId());
            }
         }

         query.setOrderBy(new int[]{table.getPrimaryKey().getId()}, false);
         query.setRange(0, 1000);
         JSONArray var16 = QueryEngine.getInstance().doQueryJSONArray(query.toSQLWithRange(), new Object[0]);
         int var17 = var16.length();
         int threshold = (int)((double)var17 * 0.6D);
         HashMap cnts = new HashMap();

         for(int colNames = 0; colNames < var16.length(); ++colNames) {
            JSONArray row = var16.getJSONArray(colNames);

            for(int col = 0; col < row.length(); ++col) {
               String cnt = ((Column)selections.get(col)).getName().toLowerCase();
               String value = row.optString(col);
               if(value != null && Validator.isNotNull(value.toString())) {
                  Integer cnt1 = (Integer)cnts.get(cnt);
                  if(cnt1 == null) {
                     cnt1 = Integer.valueOf(1);
                     cnts.put(cnt, cnt1);
                  } else {
                     cnts.put(cnt, Integer.valueOf(cnt1.intValue() + 1));
                  }
               }
            }
         }

         ArrayList var18 = new ArrayList();
         Iterator var19 = cnts.keySet().iterator();

         while(var19.hasNext()) {
            String var20 = (String)var19.next();
            int var21 = ((Integer)cnts.get(var20)).intValue();
            if(var21 >= threshold) {
               var18.add(var20);
            }

            if(var18.size() > 10) {
               break;
            }
         }

         return var18;
      }
   }

   private void backupFair() throws Exception {
      this.backupSQL("tc_fair", "select id,name,description,datebegin,dateend from b_fair where isactive=\'Y\'", "fair.csv");
   }

   private int backupSQL(String adSQLName, String sql, String csvFileName) throws Exception {
      Connection conn = null;

      int var12;
      try {
         String filePath = ConfigValues.DIR_NEA_ROOT + "/tc/" + csvFileName;
         conn = QueryEngine.getInstance().getConnection();
         String adSQL = this.getAdSQL(adSQLName, conn);
         if(Validator.isNotNull(adSQL)) {
            sql = adSQL;
         }

         long beginTime = System.currentTimeMillis();
         this.logger.debug(sql);
         int count = this.queryToFile(sql, filePath, conn);
         long duration = System.currentTimeMillis() - beginTime;
         this.logger.debug("Backup to " + filePath + " with " + count + " records using " + duration + " ms, average " + (double)count / 10.0D / (double)duration + " w/seconds");
         var12 = count;
      } finally {
         try {
            if(conn != null) {
               conn.close();
            }
         } catch (Throwable var19) {
            ;
         }

      }

      return var12;
   }

   private String status() throws Exception {
      SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
      return backupInstance != null?(backupInstance.isRunning()?"TimeCapsule is running since " + fmt.format(new Date(backupInstance.beginTime)) + ", success times: " + backupInstance.successCount + ", failed times: " + backupInstance.failedCount:"TimeCapsule is stopped, last running time:" + fmt.format(new Date(backupInstance.beginTime)) + " to " + fmt.format(new Date(backupInstance.endTime)) + ", success times: " + backupInstance.successCount + ", failed times: " + backupInstance.failedCount):"TimeCapsule is not running";
   }

   public JSONObject process(String args, JSONObject envObj) throws Exception {
      String[] arg = args.split(" ");
      IntHolder intervalHolder = new IntHolder();
      StringHolder cmdHolder = new StringHolder();
      ArgParser parser = new ArgParser("");
      parser.addOption("-c %s #cmd name", cmdHolder);
      parser.addOption("-t %d #interval", intervalHolder);
      parser.matchAllArgs(arg, 0, 0);
      String cmd = cmdHolder.value;
      int interval = intervalHolder.value;
      if(interval <= 0) {
         interval = 5;
      }

      if(Validator.isNull(cmd)) {
         return this.errorObj("Please provide sub command using -c argument: " + args);
      } else {
         String message = null;
         if("start".equalsIgnoreCase(cmd)) {
            message = this.startService(interval);
         } else if("stop".equalsIgnoreCase(cmd)) {
            message = this.stopService(true);
         } else {
            if("done".equalsIgnoreCase(cmd)) {
               HttpServletRequest jo1 = (HttpServletRequest)envObj.get("request");
               return this.done(jo1);
            }

            if("status".equalsIgnoreCase(cmd)) {
               message = this.status();
            } else {
               message = "Invalid cmd: " + cmd;
            }
         }

         JSONObject jo = null;
         jo = new JSONObject();
         jo.put("code", 0);
         jo.put("message", message);
         return jo;
      }
   }

   private void zipFolder(String inputDirectory, String zipFile) throws IOException {
      File dir = new File(inputDirectory);
      File outputZip = new File(zipFile);
      ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(outputZip));

      try {
         this.zipFolder((File)dir, (ZipOutputStream)zipOutputStream);
      } finally {
         zipOutputStream.close();
      }

   }

   private void zipFile(String sourceFile, String zipFile) throws IOException {
      File outputZip = new File(zipFile);
      ZipOutputStream zipOutputStream = new ZipOutputStream(new FileOutputStream(outputZip));

      try {
         File file = new File(sourceFile);
         String zipFilePath = file.getName();
         ZipEntry zipEntry = new ZipEntry(zipFilePath);
         zipOutputStream.putNextEntry(zipEntry);
         FileInputStream inputStream = new FileInputStream(file);
         byte[] bytes = new byte[1024];

         int length;
         while((length = inputStream.read(bytes)) >= 0) {
            zipOutputStream.write(bytes, 0, length);
         }

         zipOutputStream.closeEntry();
         inputStream.close();
      } finally {
         zipOutputStream.close();
      }
   }

   private void zipFolder(File inputDirectory, ZipOutputStream zipOutputStream) throws IOException {
      ArrayList listFiles = new ArrayList();
      this.listFiles(listFiles, inputDirectory);
      Iterator var4 = listFiles.iterator();

      while(var4.hasNext()) {
         File file = (File)var4.next();
         String filePath = file.getCanonicalPath();
         int lengthDirectoryPath = inputDirectory.getCanonicalPath().length();
         int lengthFilePath = file.getCanonicalPath().length();
         String zipFilePath = filePath.substring(lengthDirectoryPath + 1, lengthFilePath);
         ZipEntry zipEntry = new ZipEntry(zipFilePath);
         zipOutputStream.putNextEntry(zipEntry);
         FileInputStream inputStream = new FileInputStream(file);
         byte[] bytes = new byte[1024];

         int length;
         while((length = inputStream.read(bytes)) >= 0) {
            zipOutputStream.write(bytes, 0, length);
         }

         zipOutputStream.closeEntry();
         inputStream.close();
      }

   }

   private List listFiles(List listFiles, File inputDirectory) throws IOException {
      File[] allFiles = inputDirectory.listFiles();
      File[] var4 = allFiles;
      int var5 = allFiles.length;

      for(int var6 = 0; var6 < var5; ++var6) {
         File file = var4[var6];
         if(file.isDirectory()) {
            this.listFiles(listFiles, file);
         } else {
            listFiles.add(file);
         }
      }

      return listFiles;
   }

   private int queryToFile(String sql, String filePath, Connection conn) throws Exception {
      ResultSet rs = null;
      File file = (new File(filePath)).getParentFile();
      if(!file.exists()) {
         file.mkdirs();
      }

      int count = 0;
      CsvWriter writer = new CsvWriter(new PrintWriter(new OutputStreamWriter(new FileOutputStream(filePath), Charset.forName("UTF-8"))), ',');
      writer.setEscapeMode(1);
      writer.setForceQualifier(true);
      writer.setTextQualifier('\'');

      try {
         rs = conn.createStatement().executeQuery(sql);

         for(int colcnt = rs.getMetaData().getColumnCount(); rs.next(); ++count) {
            for(int i = 1; i <= colcnt; ++i) {
               Object obj = rs.getObject(i);
               if(rs.wasNull()) {
                  writer.write("NULL");
               } else {
                  writer.write(this.convertClobOrDate(obj).toString());
               }
            }

            writer.endRecord();
         }
      } finally {
         try {
            rs.close();
         } catch (Throwable var16) {
            ;
         }

         writer.flush();
         writer.close();
      }

      return count;
   }

   private Object convertClobOrDate(Object obj) throws SQLException {
      if(obj instanceof Clob) {
         obj = ((Clob)obj).getSubString(1L, (int)((Clob)obj).length());
      } else if(obj instanceof Date) {
         obj = new QDate(((Date)obj).getTime());
      }

      return obj;
   }

   public String getHelp(Locale locale) {
      return "tc -c start|stop|status|done -t [间隔分钟数] 时间胶囊TimeCapsule, 订单明细定时备份";
   }

   public String getAlias() {
      return "tc";
   }

   public void pluginDestroyed() {
      try {
         String ex = this.stopService(false);
         this.logger.debug("Stop TimeCapsule as pluginDestroyed: " + ex);
      } catch (Throwable var2) {
         this.logger.error("Fail to stop TimeCapsule service when pluginDestroyed", var2);
      }

   }

   public void pluginLoaded() {
      int interval = this.getRunningMark();
      if(interval > 0) {
         try {
            String e = this.startService(interval);
            this.logger.debug("Start TimeCapsule as pluginLoaded: " + e);
         } catch (Exception var3) {
            this.logger.error("Fail to start TimeCapsule service when pluginLoaded", var3);
         }
      }

   }


   class Backup implements Runnable {

      private int interval;
      private boolean stop = false;
      boolean isRunning;
      long beginTime;
      long endTime;
      int failedCount = 0;
      int successCount = 0;
      private long lastModifiedTime = 0L;
      boolean forceBackupEveryTime = false;


      public Backup(int interval, boolean forceBackupEveryTime) {
         this.interval = interval;
         this.forceBackupEveryTime = forceBackupEveryTime;
      }

      public void stop() {
         this.stop = true;
      }

      public boolean isRunning() {
         return this.isRunning;
      }

      private void backup() {
         try {
            File t = new File(ConfigValues.DIR_NEA_ROOT + "/tc");
            if(!t.exists()) {
               t.mkdirs();
            }

            String fileName = (new SimpleDateFormat("yyyyMMddHHmm")).format(new Date()) + ".csv";
            String filePath = ConfigValues.DIR_NEA_ROOT + "/tc/" + fileName;
            boolean shouldCreate = this.forceBackupEveryTime;
            long currentLastModified = 0L;
            if(!this.forceBackupEveryTime) {
               String lmdate = QueryEngine.getInstance().doQueryString("select to_char(max(modifieddate),\'YYYYMMDDHH24MISS\') from b_foitem", new Object[0]);
               currentLastModified = Tools.getLong(lmdate, 0L);
               shouldCreate = currentLastModified > this.lastModifiedTime;
            }

            if(shouldCreate) {
               TimeCapsule.this.backupSQL("tc_foitem", "select to_char(sysdate,\'YYYY-MM-DD HH24:MI:SS\')||\'.000\' ts, i.m_product_id pdtid, fu.user_id userid,sum(i.qty) qty, sum(i.amt_list) amtlist from b_foitem i, b_fo fo, b_funit fu where i.b_fo_id=fo.id and fu.id=fo.b_funit_id and fu.fotype=\'STANDARD\' group by i.m_product_id, fu.user_id", fileName);
               TimeCapsule.this.zipFile(filePath, filePath + ".zip");
               if(!(new File(filePath)).delete()) {
                  TimeCapsule.this.logger.warn("Could not delete " + filePath);
               }

               ++this.successCount;
               this.lastModifiedTime = currentLastModified;
            } else {
               TimeCapsule.this.logger.debug("No change for foitem since " + this.lastModifiedTime + ", skip backup");
            }
         } catch (Throwable var8) {
            TimeCapsule.this.logger.error("Fail to backup", var8);
            ++this.failedCount;
         }

      }

      public void wakeUp() {
         synchronized(this) {
            this.notifyAll();
         }
      }

      public void run() {
         this.isRunning = true;
         this.beginTime = System.currentTimeMillis();
         this.endTime = -1L;

         while(!this.stop) {
            this.backup();
            synchronized(this) {
               try {
                  this.wait((long)(this.interval * 1000 * 60));
               } catch (InterruptedException var4) {
                  TimeCapsule.this.logger.debug("Interrupted for backup");
               }
            }
         }

         TimeCapsule.this.logger.debug("Stopped for backup");
         this.isRunning = false;
         this.endTime = System.currentTimeMillis();
      }
   }
}
